
package com.openexchange.guard.hkpclient.services;

import java.util.Collection;
import java.util.List;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.openexchange.exception.OXException;
import com.openexchange.guard.exceptions.GuardCoreExceptionCodes;
import com.openexchange.guard.hkpclient.osgi.Services;
import com.openexchange.guard.keymanagement.storage.RemoteKeyCacheStorage;

/**
 *
 * {@link CachingHKPClientService} Service for querying guard's remote key cache
 *
 * @author <a href="mailto:benjamin.gruedelbach@open-xchange.com">Benjamin Gruedelbach</a>
 * @since v2.4.0
 */
public class CachingHKPClientService implements HKPClientService {

    private static final Logger logger = LoggerFactory.getLogger(CachingHKPClientService.class);
    private final HKPClientService delegate;

    /**
     * Initializes a new {@link CachingHKPClientService}.
     */
    public CachingHKPClientService() {
        this(null);
    }

    /**
     * Initializes a new {@link CachingHKPClientService}.
     *
     * @param delegate The service to delegate search queries to in a public key ring was not found
     */
    public CachingHKPClientService(HKPClientService delegate) {
        this.delegate = delegate;
    }

    /**
     * Internal method to query the OX Guard remote key cache for the given email
     *
     * @param email The email to query
     * @return The key for the given email from the cache, or null if no such key was found in the cache
     * @throws Exception
     */
    private Collection<PGPPublicKeyRing> findInCache(String email) throws Exception {
        RemoteKeyCacheStorage remoteKeyCacheStorage = Services.getService(RemoteKeyCacheStorage.class);
        List<PGPPublicKeyRing> rings = remoteKeyCacheStorage.getByEmail(email);
        return rings;
    }

    private PGPPublicKeyRing findInCache(Long id) throws Exception {
        RemoteKeyCacheStorage remoteKeyCacheStorage = Services.getService(RemoteKeyCacheStorage.class);
        PGPPublicKeyRing ring = remoteKeyCacheStorage.getByID(id);
        return ring;
    }

    /**
     * Internal method to store one or more rings in the cache for later re-use
     *
     * @param rings the key rings to save to the cache
     * @throws OXException
     */
    private void writeToCache(PGPPublicKeyRing... rings) throws OXException {
        RemoteKeyCacheStorage remoteKeyCacheStorage = Services.getService(RemoteKeyCacheStorage.class);
        for (PGPPublicKeyRing ring : rings) {
            remoteKeyCacheStorage.insert(ring);
        }
    }

    /**
     * Internal method for searching for public keys within the cache or delegate if not found
     *
     * @param clientToken an identification token put into the X-UI-INTERNAL-ACCOUNT-ID header, or null not not set the header
     * @param email The email to get the public key rings for
     * @return A list for key rings found for the given email
     * @throws Exception
     */
    private Collection<RemoteKeyResult> findAllInternal(String clientToken, String email) throws Exception {

        //Searching keys in the cache
        Collection<PGPPublicKeyRing> ret = null;
        try {
            ret = findInCache(email);
            if (ret != null && ret.size() > 0) {
                logger.debug("Remote keys found in cache");
                return RemoteKeyResult.createCollectionFrom(ret, HKPKeySources.CACHE_HKP_KEYSOURCE);
            }
        } catch (Exception e) {
            logger.error("Error querying HKP key cache", e);
        }

        Collection<RemoteKeyResult> result = null;
        //Search delegate
        if (delegate != null) {
            result = delegate.find(clientToken, email);
        }

        if (result != null && !result.isEmpty()) {
            //Only save from trusted sources
            Collection<PGPPublicKeyRing> trustedResults = RemoteKeyResult.getRingsFrom(RemoteKeyResult.getTrusted(result));
            if(trustedResults != null && trustedResults.size() > 0) {
                logger.debug("Writing {} found keys from trusted HKP server into cache", trustedResults.size());
                writeToCache(trustedResults.toArray(new PGPPublicKeyRing[trustedResults.size()]));
            }
        }
        return result;
    }

    /**
     * Search for a key by long ID in the remote key cache
     *
     * @param clientToken
     * @param id
     * @param email
     * @return
     * @throws Exception
     */
    private RemoteKeyResult findInternal(String clientToken, Long id, String email) throws Exception {

        PGPPublicKeyRing ring = findInCache(id);
        if (ring != null) {
            return new RemoteKeyResult(ring, HKPKeySources.CACHE_HKP_KEYSOURCE);
        }

        RemoteKeyResult result = null;
        //Search delegate
        if (delegate != null) {
            result = delegate.find(clientToken, email, id);
        }

        if (result != null && result.hasResult()) {
            if (result.getSource().isTrusted()) { // Only save from trusted sources
                logger.debug("Writing found keys from HKP server into cache");
                writeToCache(result.getRing());
            }
        }
        return result;
    }

    @Override
    public Collection<RemoteKeyResult> find(String clientToken, String email) throws OXException {
        try {
            return findAllInternal(clientToken, email);
        } catch (Exception e) {
            throw GuardCoreExceptionCodes.IO_ERROR.create(e, e.getMessage());
        }
    }

    @Override
    public RemoteKeyResult find(String clientToken, String email, Long id) throws OXException {
        try {
            return findInternal(clientToken, id, email);
        } catch (Exception e) {
            throw GuardCoreExceptionCodes.IO_ERROR.create(e, e.getMessage());
        }
    }
}
